/* Nihongo Benkyo
* Copyright (C) 2005 Mathieu Blondel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <sqlite3.h>
#include "nihongobenkyo_import.h"

static GIOChannel *
edict_open(char *file) {
    return g_io_channel_new_file(file, "r", NULL);
}

static int 
edict_read(GIOChannel *edict_file, GString *line) {
    GIOStatus status;
    
    status = g_io_channel_read_line_string(edict_file, line, NULL, NULL);
    
    if(status == G_IO_STATUS_EOF)
        return 0;
    else
        return 1;
}

static gchar *
edict_get_kanji_and_kana(GString *line, GString *kana, GString *kanji) {
    /* Gets kanji and kana and return translations positions
    of a line with the following format :
    
    KANJI [KANA] /English_1/English_2/.../
    or
    KANA /English_1/.../
    */
    gchar *current_char = line->str;
    GString *temp = g_string_new(NULL);
    g_string_assign(kana, "");
    g_string_assign(kanji, "");
    
    /* The line must be a valid utf8 string */
    if(g_utf8_validate(line->str, -1, NULL)) {
        
        /* Kana or Kanji */
        
        while(*current_char != ' ') {
            g_string_append_unichar(temp, g_utf8_get_char(current_char));
            current_char = g_utf8_next_char(current_char);
        }
        
        /* Skips the white space after the kanji or kana */
        current_char = g_utf8_next_char(current_char);    
        
        /* Kana if Kanji */
        if(*current_char == '[') {
        
            /* The temp string was in fact a Kanji */
            *kanji = *temp;
            
            /* Skips '[' */
            current_char = g_utf8_next_char(current_char);
            
            while(*current_char != ']') {
                g_string_append_unichar(kana, g_utf8_get_char(current_char));
                current_char = g_utf8_next_char(current_char);       
            }
            
            /* Skips ']' and the white space after the kana */
            current_char = g_utf8_next_char(current_char); 
            current_char = g_utf8_next_char(current_char);  
            
        }
        else {   
            /* The temp string was in fact a Kana */
            *kana = *temp;
        }
        
        return current_char;    
   
    }
    else {
        rb_raise (rb_eDicTypeException, "Is not encoded in UTF-8");
    }
}

static gchar* 
edict_get_translation(gchar *current_char, GString *translation) {
    g_string_assign(translation, "");
    
    if(*current_char == '/') {
        current_char = g_utf8_next_char(current_char);
        while(*current_char != '/') {
            if(*current_char == '\n') {
                return NULL;
            }
            else {
                g_string_append_unichar(translation, g_utf8_get_char(current_char));
                current_char = g_utf8_next_char(current_char);
            }
        }
        return current_char;
    }
    else {
        return NULL;
    }        
}

static void 
edict_close(GIOChannel *edict_file) {
    if(edict_file != NULL)
        g_io_channel_close(edict_file);
}

void 
edict_import(char *db_file, char *dic_file, char *dic_name) {

    /* Database pointer */
    sqlite3 *sqlite3_db;
    int db_open;

    /* Ids */
    int dic_id;
    int dic_position;
    int entry_id;
    int kanji_id;
    int reading_id;
    int translation_id;
    int sense_id;
    
    /* Variables that store extracted data */
    GIOChannel *edict_file;
    gchar *translation_position;
    GString *line = g_string_new("");
    GString *kana_reading = g_string_new("");
    GString *kanji = g_string_new("");
    GString *translation = g_string_new("");
    
    /* Open database */
    db_open = database_open(&sqlite3_db, db_file);
    if(db_open != SQLITE_OK) {
        rb_raise (rb_eIOError, "Cannot open database");
    }    
   
    /* Dic id */
    if(database_dictionary_exists(sqlite3_db, dic_name)) {
        database_close(sqlite3_db);
        rb_raise (rb_eDicAlreadyExistsException, "Dictionary already exists");   
    }
    
    if(database_has_indexes(sqlite3_db) == 1)database_drop_indexes(sqlite3_db);
    
    dic_position = database_get_last_dictionary_id(sqlite3_db);
    database_add_dictionary(sqlite3_db, dic_name, WORD_DIC, dic_position);
    dic_id = database_get_dictionary_id(sqlite3_db, dic_name);
    
    /* Latest ids found in the database */
    entry_id = database_get_last_entry_id(sqlite3_db) + 1;
    kanji_id = database_get_last_kanji_id(sqlite3_db) + 1;
    reading_id = database_get_last_reading_id(sqlite3_db) + 1;
    translation_id = database_get_last_translation_id(sqlite3_db) + 1;
    sense_id = database_get_last_sense_id(sqlite3_db) + 1;

    /* Open EDICT file */
    edict_file = edict_open(dic_file);
    
    if(edict_file == NULL) {
        rb_raise (rb_eIOError, "Could not open dic file");  
    }
    
    /* Adding entries kanji kana and translations */
    while(edict_read(edict_file, line)) {
    
        database_add_entry(sqlite3_db, dic_id, entry_id);    
    
        translation_position = edict_get_kanji_and_kana(line, kana_reading, kanji);
        database_add_kanji(sqlite3_db, entry_id, kanji_id, kanji->str);
        database_add_reading(sqlite3_db, entry_id, reading_id, "", kana_reading->str);
        
        translation_position = edict_get_translation(translation_position,translation);
                
        while(translation_position != NULL) {
            database_add_translation(sqlite3_db, entry_id, translation_id, translation->str, sense_id, "");
            translation_position = edict_get_translation(translation_position,translation); 
            translation_id++;           
        }
        
        /* IN EDICT, an entry has one kanji, one reading and one sense only */
        entry_id++;
        kanji_id++;
        reading_id++;
        sense_id++;
        
    }
    
    /* Close EDICT file */
    edict_close(edict_file);
    
    database_create_indexes(sqlite3_db);
    
    /* Close database */
    database_close(sqlite3_db);
}
