/* Nihongo Benkyo
* Copyright (C) 2005 Mathieu Blondel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <sqlite3.h>
#include "nihongobenkyo_import.h"

static sqlite3 *sqlite3_db;
static int dic_id, entry_id, kanji_id, reading_id, translation_id, sense_id;
static enum textNodeName current_text_node;
static struct jmdict_langs l;
static xmlChar *current_lang;

static int
lang_ok(xmlChar *lang) {
    int ret = 0;
    if((xmlStrcmp(lang, (xmlChar *) "en") == 0 && l.en == 1) ||
       (xmlStrcmp(lang, (xmlChar *) "fr") == 0 && l.fr == 1) ||
       (xmlStrcmp(lang, (xmlChar *) "de") == 0 && l.de == 1) ||
       (xmlStrcmp(lang, (xmlChar *) "nl") == 0 && l.nl == 1) ||
       (xmlStrcmp(lang, (xmlChar *) "ru") == 0 && l.ru == 1))
       ret = 1;
    return ret;
}

static void
process_jmdict_node(xmlTextReaderPtr reader) {
    const xmlChar *name, *value;

    name = xmlTextReaderConstName(reader);
    if (name == NULL)
        name = (xmlChar *) "";

    value = xmlTextReaderConstValue(reader);

    /* Close tags */
    
    if(end_element(reader, name, NULL)) {
        current_text_node = NONE;
    }
    
    /* Entry tags */
    
    if(start_element(reader, name, "entry")) {
        database_add_entry(sqlite3_db, dic_id, entry_id);
    }
    else if(end_element(reader, name, "entry")) {
        entry_id++;
    }  
    
    /* Kanji tags */
    
    else if(start_element(reader, name, "keb")) {
            current_text_node = KEB;
    }  

    else if(end_element(reader, name, "k_ele")) {
        kanji_id++;
    }
    else if(current_text_node == KEB) {
        database_add_kanji(sqlite3_db, entry_id, kanji_id, (char *) value);
    }
    
    /* Kana tags */
    
    else if(start_element(reader, name, "reb")) {
        current_text_node = REB;    
    }
    else if(end_element(reader, name, "r_ele")) {
        reading_id++;
    }
    else if(current_text_node == REB) {
        database_add_reading(sqlite3_db, entry_id, reading_id, "", (char *) value);
    }
    
    /* Gloss tags */
    
    else if(start_element(reader, name, "gloss")) {
        
        current_lang = xmlTextReaderGetAttribute(reader, (xmlChar *) "g_lang");
        
        if(lang_ok(current_lang))
            current_text_node = GLOSS;
    }
    else if(end_element(reader, name, "gloss")) {
            translation_id++;
    }
    else if(current_text_node == GLOSS) {
        database_add_translation(sqlite3_db, entry_id, translation_id, (char *) value,
         sense_id, (char *) current_lang);
    }
    
    /* Sense tags */
    
    else if(end_element(reader, name, "sense")) {
        sense_id++;
    }
    
    /* Metadata */

    else { 
              
        /* For kanji */

        if(start_element(reader, name, "ke_inf")) {
            current_text_node = KE_INF;
        }
        else if(current_text_node == KE_INF) {
            database_add_metadata_value(sqlite3_db, kanji_id, WORD_KANJI_INFO, (char *) value);
        }
        else if(start_element(reader, name, "ke_pri")) {
            current_text_node = KE_PRI;
        }
        else if(current_text_node == KE_PRI) {
            if(xmlStrcmp(value, (xmlChar *) "gai1") == 0 ||
               xmlStrcmp(value, (xmlChar *) "spec1") == 0 ||
               xmlStrncmp(value, (xmlChar *) "nf", 2) == 0) 
            {
               database_add_metadata_value(sqlite3_db, kanji_id, WORD_KANJI_PRIORITY, (char *) value);
            }
            else {
                database_add_metadata_value(sqlite3_db, kanji_id, WORD_KANJI_SOURCE, (char *) value);        
            }
        } 
        
        /* For readings */
        
        else if(start_element(reader, name, "re_inf")) {
            current_text_node = RE_INF;
        }
        else if(current_text_node == RE_INF) {
            database_add_metadata_value(sqlite3_db, reading_id, WORD_READING_INFO, (char *) value);
        }
        
        else if(start_element(reader, name, "re_pri")) {
            current_text_node = RE_PRI;
        }
        else if(current_text_node == RE_PRI) {
            if(xmlStrcmp(value, (xmlChar *) "gai1") == 0 ||
               xmlStrcmp(value, (xmlChar *) "spec1") == 0 ||
               xmlStrncmp(value, (xmlChar *) "nf", 2) == 0) 
            {
               database_add_metadata_value(sqlite3_db, reading_id, WORD_READING_PRIORITY, (char *) value);
            }
            else {
                database_add_metadata_value(sqlite3_db, reading_id, WORD_READING_SOURCE, (char *) value);        
            }
        }   
        else if(start_element(reader, name, "re_nokanji")) {
            database_add_metadata_value(sqlite3_db, reading_id, WORD_READING_NO_KANJI, "yes");
        }
        else if(start_element(reader, name, "re_restr")) {
            current_text_node = RE_RESTR;
        }
        else if(current_text_node == RE_RESTR) {
            database_add_metadata_value(sqlite3_db, reading_id, WORD_READING_RESTRICTION, (char *) value);
        } 
        
        /* For entries * */
        
        else if(start_element(reader, name, "lang")) {
            current_text_node = LANG;
        }
        else if(current_text_node == LANG) {
            database_add_metadata_value(sqlite3_db, entry_id, WORD_ENTRY_ORIGIN_LANG, (char *) value);
        }  
        
        else if(start_element(reader, name, "dial")) {
            current_text_node = DIAL;
        }
        else if(current_text_node == DIAL) {
            database_add_metadata_value(sqlite3_db, entry_id, WORD_ENTRY_DIALECT, (char *) value);
        } 
        
        else if(start_element(reader, name, "etym")) {
            current_text_node = ETYM;
        }
        else if(current_text_node == ETYM) {
            database_add_metadata_value(sqlite3_db, entry_id, WORD_ENTRY_ETYMOLOGY, (char *) value);
        } 
        
        /* For senses */
        
        else if(start_element(reader, name, "stagk")) {
            current_text_node = STAGK;
        }
        else if(current_text_node == STAGK) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_KANJI_RESTRICTION, (char *) value);
        }     
        
        else if(start_element(reader, name, "stagr")) {
            current_text_node = STAGR;
        }
        else if(current_text_node == STAGR) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_READING_RESTRICTION, (char *) value);
        }     
        
        else if(start_element(reader, name, "xref")) {
            current_text_node = XREF;
        }
        else if(current_text_node == XREF) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_REFERENCE, (char *) value);
        }   
        
        else if(start_element(reader, name, "ant")) {
            current_text_node = ANT;
        }
        else if(current_text_node == ANT) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_ANTONYM, (char *) value);
        }    
        
        else if(start_element(reader, name, "pos")) {
            current_text_node = POS;
        }
        else if(current_text_node == POS) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_PART_OF_SPEECH, (char *) value);
        }        
        
        else if(start_element(reader, name, "field")) {
            current_text_node = FIELD;
        }
        else if(current_text_node == FIELD) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_FIELD, (char *) value);
        }               
        
        else if(start_element(reader, name, "misc")) {
            current_text_node = MISC;
        }
        else if(current_text_node == MISC) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_MISC, (char *) value);
        }     
        
        else if(start_element(reader, name, "s_inf")) {
            current_text_node = S_INF;
        }
        else if(current_text_node == S_INF) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_INFO, (char *) value);
        }
   
   }
    
}

void 
jmdict_import(char *db_file, char* dic_file, char *dic_name, struct jmdict_langs langs)
{
    int db_open;
    int dic_position; 
    
    /* Open database */
    db_open = database_open(&sqlite3_db, db_file);
    if(db_open != SQLITE_OK) {
        rb_raise (rb_eIOError, "Cannot open database");
    }    
   
    /* Dic id */
    if(database_dictionary_exists(sqlite3_db, dic_name)) {
        database_close(sqlite3_db);
        rb_raise (rb_eDicAlreadyExistsException, "Dictionary already exists");   
    }
    
    if(database_has_indexes(sqlite3_db) == 1)database_drop_indexes(sqlite3_db);
    
    dic_position = database_get_last_dictionary_id(sqlite3_db);
    database_add_dictionary(sqlite3_db, dic_name, WORD_DIC, dic_position);
    dic_id = database_get_dictionary_id(sqlite3_db, dic_name);

    /* Latest ids found in the database */
    entry_id = database_get_last_entry_id(sqlite3_db) + 1;
    kanji_id = database_get_last_kanji_id(sqlite3_db) + 1;
    reading_id = database_get_last_reading_id(sqlite3_db) + 1;
    translation_id = database_get_last_translation_id(sqlite3_db) + 1;
    sense_id = database_get_last_sense_id(sqlite3_db) + 1;
    l = langs;    
    
    /* Begin parsing */
    stream_file(dic_file, &process_jmdict_node);
    
    database_create_indexes(sqlite3_db);
    
    database_close(sqlite3_db);
}
