/* Nihongo Benkyo
* Copyright (C) 2005 Mathieu Blondel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <sqlite3.h>
#include "nihongobenkyo_import.h"

static sqlite3 *sqlite3_db;
static int dic_id, entry_id, kanji_id, reading_id, translation_id, sense_id;
static enum textNodeName current_text_node;

static void
process_jmnedict_node(xmlTextReaderPtr reader) {
    const xmlChar *name, *value;

    name = xmlTextReaderConstName(reader);
    if (name == NULL)
        name = (xmlChar *) "";

    value = xmlTextReaderConstValue(reader);

    /* Close tags */
    
    if(end_element(reader, name, NULL)) {
        current_text_node = NONE;
    }
    
    /* Entry tags */
    
    if(start_element(reader, name, "entry")) {
        database_add_entry(sqlite3_db, dic_id, entry_id);
    }
    else if(end_element(reader, name, "entry")) {
        entry_id++;
        sense_id++;
    }  
    
    /* Kanji tags */
    
    else if(start_element(reader, name, "keb")) {
            current_text_node = KEB;
    }  
    else if(end_element(reader, name, "k_ele")) {
            kanji_id++;
    }
    else if(current_text_node == KEB) {
        database_add_kanji(sqlite3_db, entry_id, kanji_id, (char *) value);
    }
    
    /* Kana tags */
    
    else if(start_element(reader, name, "reb")) {
        current_text_node = REB;    
    }
    else if(end_element(reader, name, "r_ele")) {
            reading_id++;
    }
    else if(current_text_node == REB) {
        database_add_reading(sqlite3_db, entry_id, reading_id, "", (char *) value);
    }
    
    /* trans_det tags */
    
    else if(start_element(reader, name, "trans_det")) {     
        current_text_node = TRANS_DET;
    }
    else if(end_element(reader, name, "trans_det")) {
        translation_id++;
    }
    else if(current_text_node == TRANS_DET) {
        database_add_translation(sqlite3_db, entry_id, translation_id, (char *) value, sense_id, "");
    }
    
    /* Metadata */

    else { 
        
        /* For kanji */

        if(start_element(reader, name, "name_type")) {
            current_text_node = NAME_TYPE;
        }
        else if(current_text_node == NAME_TYPE) {
            database_add_metadata_value(sqlite3_db, sense_id, 
                                        WORD_SENSE_NAME_TYPE, (char *) value);
        }
        
    }

    
}

void 
jmnedict_import(char *db_file, char *dic_file, char *dic_name)
{
    int db_open;
    int dic_position; 
    
    /* Open database */
    db_open = database_open(&sqlite3_db, db_file);
    if(db_open != SQLITE_OK) {
        rb_raise (rb_eIOError, "Cannot open database");
    }    
   
    /* Dic id */
    if(database_dictionary_exists(sqlite3_db, dic_name)) {
        database_close(sqlite3_db);
        rb_raise (rb_eDicAlreadyExistsException, "Dictionary already exists");   
    }
    
    if(database_has_indexes(sqlite3_db) == 1)database_drop_indexes(sqlite3_db);
    
    dic_position = database_get_last_dictionary_id(sqlite3_db);
    database_add_dictionary(sqlite3_db, dic_name, WORD_DIC, dic_position);
    dic_id = database_get_dictionary_id(sqlite3_db, dic_name);

    /* Latest ids found in the database */
    entry_id = database_get_last_entry_id(sqlite3_db) + 1;
    kanji_id = database_get_last_kanji_id(sqlite3_db) + 1;
    reading_id = database_get_last_reading_id(sqlite3_db) + 1;
    translation_id = database_get_last_translation_id(sqlite3_db) + 1;
    sense_id = database_get_last_sense_id(sqlite3_db) + 1;  
    
    /* Begin parsing */
    stream_file(dic_file, &process_jmnedict_node);
    
    database_create_indexes(sqlite3_db);
    
    database_close(sqlite3_db);
}
