/* Nihongo Benkyo
* Copyright (C) 2005 Mathieu Blondel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <sqlite3.h>
#include "nihongobenkyo_import.h"

static sqlite3 *sqlite3_db;
static int dic_id, entry_id, kanji_id, reading_id, translation_id, sense_id;
static enum textNodeName current_text_node;
static struct kanjidic2_langs l;
static xmlChar *current_lang;
static xmlChar *current_reading_type;

static int
lang_ok(xmlChar *lang) {
    int ret = 0;
    if((xmlStrcmp(lang, (xmlChar *) "en") == 0 && l.en == 1) ||
       (xmlStrcmp(lang, (xmlChar *) "es") == 0 && l.es == 1) ||
       (xmlStrcmp(lang, (xmlChar *) "pt") == 0 && l.pt == 1))
       ret = 1;
    return ret;
}

static void
process_kanjidic2_node(xmlTextReaderPtr reader) {
    const xmlChar *name, *value;

    name = xmlTextReaderConstName(reader);
    if (name == NULL)
        name = (xmlChar *) "";

    value = xmlTextReaderConstValue(reader);

    /* Close tags */
    
    if(end_element(reader, name, NULL)) {
        current_text_node = NONE;
    }
    
    /* Character */
    
    if(start_element(reader, name, "character")) {
        database_add_entry(sqlite3_db, dic_id, entry_id);
    }
    else if(end_element(reader, name, "character")) {
        entry_id++;
    }  
    
    /* Literal tags */
    
    else if(start_element(reader, name, "literal")) {
            current_text_node = LITERAL;
    }  

    else if(end_element(reader, name, "literal")) {
        kanji_id++;
    }
    else if(current_text_node == LITERAL) {
        database_add_kanji(sqlite3_db, entry_id, kanji_id, (char *) value);
    }
    
    /* Reading tags */
    
    else if(start_element(reader, name, "reading")) {
        current_text_node = READING;    
        
        current_reading_type = xmlTextReaderGetAttribute(reader, (xmlChar *) "r_type");
    }
    else if(end_element(reader, name, "reading")) {
        reading_id++;
    }
    else if(current_text_node == READING) {
        database_add_reading(sqlite3_db, entry_id, reading_id, (char *) current_reading_type, (char *) value);
    }
    
    /* Meaning tags */
    
    else if(start_element(reader, name, "meaning")) {
        
        current_lang = xmlTextReaderGetAttribute(reader, (xmlChar *) "m_lang");
        
        if(current_lang == NULL)current_lang = (xmlChar *) "en";
        
        if(lang_ok(current_lang))
            current_text_node = MEANING;
    }
    else if(end_element(reader, name, "meaning")) {
            translation_id++;
    }
    else if(current_text_node == MEANING) {
        database_add_translation(sqlite3_db, entry_id, translation_id, (char *) value,
         sense_id, (char *) current_lang);
    }
    
    /* Rmgroup tags */
    
    else if(end_element(reader, name, "rmgroup")) {
        sense_id++;
    }    
    
    /* Grade tags */
    
    else if(start_element(reader, name, "grade")) {
        current_text_node = GRADE;
    }
    else if(current_text_node == GRADE) {
        database_add_metadata_value(sqlite3_db, entry_id, 
                                    KANJI_GRADE, (char *) value);
    }
    
    /* Stroke tags */
    
    else if(start_element(reader, name, "stroke_count")) {
        current_text_node = STROKE_COUNT;
    }
    else if(current_text_node == STROKE_COUNT) {
        database_add_metadata_value(sqlite3_db, entry_id, 
                                    KANJI_STROKE_COUNT, (char *) value);
    }
    
    /* Frequence tags */
    
    else if(start_element(reader, name, "freq")) {
        current_text_node = FREQ;
    }
    else if(current_text_node == FREQ) {
        database_add_metadata_value(sqlite3_db, entry_id, 
                                    KANJI_FREQUENCE, (char *) value);
    }
    
}

void 
kanjidic2_import(char *db_file, char* dic_file, char *dic_name, struct kanjidic2_langs langs)
{
    int db_open;
    int dic_position; 
    
    /* Open database */
    db_open = database_open(&sqlite3_db, db_file);
    if(db_open != SQLITE_OK) {
        rb_raise (rb_eIOError, "Cannot open database");
    }    
   
    /* Dic id */
    if(database_dictionary_exists(sqlite3_db, dic_name)) {
        database_close(sqlite3_db);
        rb_raise (rb_eDicAlreadyExistsException, "Dictionary already exists");   
    }
    
    if(database_has_indexes(sqlite3_db) == 1)database_drop_indexes(sqlite3_db);
    
    dic_position = database_get_last_dictionary_id(sqlite3_db);
    database_add_dictionary(sqlite3_db, dic_name, KANJI_DIC, dic_position);
    dic_id = database_get_dictionary_id(sqlite3_db, dic_name);

    /* Latest ids found in the database */
    entry_id = database_get_last_entry_id(sqlite3_db) + 1;
    kanji_id = database_get_last_kanji_id(sqlite3_db) + 1;
    reading_id = database_get_last_reading_id(sqlite3_db) + 1;
    translation_id = database_get_last_translation_id(sqlite3_db) + 1;
    sense_id = database_get_last_sense_id(sqlite3_db) + 1;
    l = langs;    
    
    /* Begin parsing */
    stream_file(dic_file, &process_kanjidic2_node);
    
    database_create_indexes(sqlite3_db);
    
    database_close(sqlite3_db);
}
