/* Nihongo Benkyo
* Copyright (C) 2005 Mathieu Blondel
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sqlite3.h>
#include <glib.h>

static sqlite3_stmt *add_dic_statement;
static sqlite3_stmt *add_entry_statement;
static sqlite3_stmt *add_translation_statement;
static sqlite3_stmt *add_kanji_statement;
static sqlite3_stmt *add_reading_statement;
static sqlite3_stmt *add_metadata_statement;

int 
database_execute(sqlite3 *db, char *query);

int 
database_get_first_int_value_callback(void *p_data, int num_fields, char **p_fields, char **p_col_names) {
    int *ptr = (int *)p_data;
    
    if(num_fields >= 1 && p_fields[0])
        *ptr = atoi(p_fields[0]);
    else *ptr = 0;
    
    return 0;
}

int 
database_exec(sqlite3 *db, char *query, sqlite3_callback callback, void *callback_first_argument) {
    /* Same as sqlite3_exec but with error handling and begin-commit */
    int result;
    char *error_msg;
        
    result = sqlite3_exec(db, query, callback, callback_first_argument, &error_msg);
    
    if(result != SQLITE_OK) {
        fprintf(stderr, "SQL error: %s\n", error_msg);
    }
        
    return result;
}

int 
database_execute(sqlite3 *db, char *query) {
    /* To use when we do not need a response : INSERT, UPDATE, DELETE statements */
    return database_exec(db, query, NULL, NULL);
}

int 
database_bind_text(sqlite3_stmt *statement, int index, char *text) {
    int r;
    
    r = sqlite3_bind_text (statement, index, text, -1, SQLITE_STATIC);
    if(r != SQLITE_OK) {
        fprintf(stderr, "Could not bind text\n");
    }
        
    return r;       
}

int 
database_bind_int(sqlite3_stmt *statement, int index, int value) {
    int r;
    
    r= sqlite3_bind_int(statement, index, value);
    if(r != SQLITE_OK) {
        fprintf(stderr, "Could not bind int.\n");
    }
    
    return r;
}

int 
database_step(sqlite3_stmt *statement) {
    int r;

    r = sqlite3_step(statement);
    if(r != SQLITE_DONE) {
        fprintf(stderr, "Could not insert translation \n");
        fprintf(stderr, "%d", r);
    }
        
    return r;
}

int 
database_reset(sqlite3_stmt *statement) {
    return sqlite3_reset(statement);
}

int 
database_step_and_reset(sqlite3_stmt *statement) {
    database_step(statement);
    return database_reset(statement);
}

int 
database_prepare(sqlite3 *db, char *query, sqlite3_stmt **statement) {
    int r;
    r = sqlite3_prepare(db, query, -1, statement, NULL);
    if(r!= SQLITE_OK)
        fprintf(stderr, "Could not prepare statement");
    return r;
}

int 
database_open(sqlite3 **db, char *database_file) {
    int result;
    
    result = sqlite3_open(database_file, db);
    
    if(result == SQLITE_OK) {
        database_execute(*db, "BEGIN");
    }
        
    return result;
}

int 
database_close(sqlite3 *db) {
    database_execute(db, "COMMIT");
    return sqlite3_close(db);
}

int 
database_add_dictionary(sqlite3 *db, char *dic_name, int dic_type, int position) {
    add_dic_statement = NULL;
    add_entry_statement = NULL;
    add_translation_statement = NULL;
    add_kanji_statement = NULL;
    add_reading_statement = NULL;
    add_metadata_statement = NULL;    

    database_prepare(db, 
    "INSERT INTO dictionaries (id,name,type,position) VALUES (NULL,?,?,?);",
    &add_dic_statement);
    
    database_bind_text(add_dic_statement, 1, dic_name);
    database_bind_int(add_dic_statement, 2, dic_type);
    database_bind_int(add_dic_statement, 3, position);
    return database_step_and_reset(add_dic_statement);
}

int 
database_add_entry(sqlite3 *db, int dictionary_id, int entry_id) {
    
    if(add_entry_statement == NULL)
        database_prepare(db, 
                         "INSERT INTO entries (id,dictionary_id) VALUES (?, ?)",
                         &add_entry_statement);
    
    database_bind_int(add_entry_statement, 1, entry_id);
    database_bind_int(add_entry_statement, 2, dictionary_id);
    return database_step_and_reset(add_entry_statement);
}

int 
database_add_kanji(sqlite3 *db, int entry_id, int kanji_id, char *kanji) {
    
    if(add_kanji_statement == NULL)
    database_prepare(db, 
    "INSERT INTO kanji (id,entry_id,kanji) VALUES (?,?,?);",
    &add_kanji_statement);

    database_bind_int(add_kanji_statement, 1, kanji_id);
    database_bind_int(add_kanji_statement, 2, entry_id);
    database_bind_text(add_kanji_statement, 3, kanji);    

    return database_step_and_reset(add_kanji_statement);
}

int 
database_add_reading(sqlite3 *db, int entry_id, int reading_id, char *type, char *reading) {

    if(add_reading_statement == NULL)
    database_prepare(db, 
    "INSERT INTO readings (id,entry_id,type,reading) VALUES (?,?,?,?);",
    &add_reading_statement);
    
    database_bind_int(add_reading_statement, 1, reading_id);
    database_bind_int(add_reading_statement, 2, entry_id);
    database_bind_text(add_reading_statement, 3, type);
    database_bind_text(add_reading_statement, 4, reading);    
    return database_step_and_reset(add_reading_statement); 
}

int 
database_add_translation(sqlite3 *db, int entry_id, int translation_id, 
                              char *translation, int sense_id, char *language) {  
    
    if(add_translation_statement == NULL)
    database_prepare(db, 
    "INSERT INTO translations (id,entry_id,sense_id,language,translation) VALUES (?,?,?,?,?);",
    &add_translation_statement);
    
    database_bind_int(add_translation_statement, 1, translation_id);
    database_bind_int(add_translation_statement, 2, entry_id);
    database_bind_int(add_translation_statement, 3, sense_id);
    database_bind_text(add_translation_statement, 4, language);  
    database_bind_text(add_translation_statement, 5, translation);    
    return database_step_and_reset(add_translation_statement); 
}

int 
database_add_metadata_value(sqlite3 *db, int matching_id, int metadata_id, char *metadata_value) {  
    
    if(add_metadata_statement == NULL)
    database_prepare(db, 
    "INSERT INTO metadata_values (id,matching_id,metadata_name_id,value) VALUES (NULL,?,?,?);",
    &add_metadata_statement);
    
    database_bind_int(add_metadata_statement, 1, matching_id);
    database_bind_int(add_metadata_statement, 2, metadata_id);
    database_bind_text(add_metadata_statement, 3, metadata_value);
    return database_step_and_reset(add_metadata_statement); 
}

int 
database_get_last_entry_id(sqlite3 *db) {
    int n = 0;
    database_exec(db, "SELECT MAX(id) FROM entries", 
    database_get_first_int_value_callback, &n);
    return n;
}

int 
database_get_last_kanji_id(sqlite3 *db) {
    int n = 0;
    database_exec(db, "SELECT MAX(id) FROM kanji", 
    database_get_first_int_value_callback, &n);
    return n;
}

int 
database_get_last_reading_id(sqlite3 *db) {
    int n = 0;
    database_exec(db, "SELECT MAX(id) FROM readings", 
    database_get_first_int_value_callback, &n);
    return n;
}

int 
database_get_last_translation_id(sqlite3 *db) {
    int n = 0;
    database_exec(db, "SELECT MAX(id) FROM translations", 
    database_get_first_int_value_callback, &n);
    return n;
}

int 
database_get_last_sense_id(sqlite3 *db) {
    int n = 0;
    database_exec(db, "SELECT MAX(sense_id) FROM translations", 
    database_get_first_int_value_callback, &n);
    return n;
}


int 
database_get_last_dictionary_id(sqlite3 *db) {
    int n = 0;
    database_exec(db, "SELECT MAX(id) FROM dictionaries", 
    database_get_first_int_value_callback, &n);
    return n;
}

int 
database_get_dictionary_id(sqlite3 *db, char *dic_name) {
    int n;
    GString *query = g_string_new(NULL);
    g_string_printf(query, "SELECT id FROM dictionaries WHERE name = '%s' LIMIT 1", dic_name);
    
    database_exec(db, query->str, 
    database_get_first_int_value_callback, &n);
    return n; 
}

int 
database_dictionary_exists(sqlite3 *db, char *dic_name) {
    int n;
    GString *query = g_string_new(NULL);
    g_string_printf(query, "SELECT count(id) FROM dictionaries WHERE name = '%s'", dic_name);
    
    database_exec(db, query->str, 
    database_get_first_int_value_callback, &n);
    if(n >= 1)return 1;
    else return 0; 
}

void 
database_create_indexes(sqlite3 *db) {
    database_execute(db, "CREATE INDEX kanji_entry_index ON kanji (entry_id);");
    database_execute(db, "CREATE INDEX kanji_index ON kanji (kanji);");
    
    database_execute(db, "CREATE INDEX reading_entry_index ON readings (entry_id);");
    database_execute(db, "CREATE INDEX reading_index ON readings (reading);");
    
    database_execute(db, "CREATE INDEX translation_entry_index ON translations (entry_id);");
    database_execute(db, "CREATE INDEX translation_index ON translations (translation);");
    
    database_execute(db, "CREATE INDEX metadata_value_matching_id_index ON metadata_values (matching_id);");
    /*database_execute(db, "CREATE INDEX metadata_value_index ON metadata_values (value);");*/
}

void 
database_drop_indexes(sqlite3 *db) {
    database_execute(db, "DROP INDEX kanji_entry_index;");
    database_execute(db, "DROP INDEX kanji_index;");
    
    database_execute(db, "DROP INDEX reading_entry_index;");
    database_execute(db, "DROP INDEX reading_index;");
    
    database_execute(db, "DROP INDEX translation_entry_index;");
    database_execute(db, "DROP INDEX translation_index;");
    
    database_execute(db, "DROP INDEX metadata_value_matching_id_index;");
    /*database_execute(db, "DROP INDEX metadata_value_index;");*/
}

int 
database_has_indexes(sqlite3 *db) {
    int n;

    database_exec(db, "SELECT count(*) FROM sqlite_master WHERE type = 'index'", 
    database_get_first_int_value_callback, &n);
    if(n >= 1)return 1;
    else return 0; 
}
